const char hb_builtin_presets_json[] =
"{\n"
"    \"PresetBuiltin\": [\n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"fast\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"Small H.264 video (up to 1080p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Very Fast 1080p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 4000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"veryfast\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 24.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": true, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"fast\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"Small H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Very Fast 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 2000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"veryfast\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 23.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": true, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"fast\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 576, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"Small H.264 video (up to 576p25) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Very Fast 576p25\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1200, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"25\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"veryfast\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": true, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"fast\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"Small H.264 video (up to 480p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Very Fast 480p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"veryfast\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": true, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": true, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Fast 1080p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"fast\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": true, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Fast 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 3000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"fast\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": true, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 576, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"H.264 video (up to 576p25) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Fast 576p25\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1800, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"25\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"fast\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": true, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"H.264 video (up to 480p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Fast 480p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"fast\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": true, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"HQ 1080p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 9000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 720p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"HQ 720p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 4500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 19.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 576, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 576p25), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"HQ 576p25 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 2700, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"25\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 18.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 480p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"HQ 480p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 2250, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 18.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"Super high quality H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Super HQ 1080p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 15000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"ref=5:bframes=5\", \n"
"                    \"VideoPreset\": \"veryslow\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 18.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"Super high quality H.264 video (up to 720p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Super HQ 720p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 7500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"ref=5:bframes=5\", \n"
"                    \"VideoPreset\": \"veryslow\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 17.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 576, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"Super high quality H.264 video (up to 576p25), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Super HQ 576p25 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 4500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"25\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"ref=5:bframes=5\", \n"
"                    \"VideoPreset\": \"veryslow\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 16.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"Super high quality H.264 video (up to 480p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Super HQ 480p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 3750, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"ref=5:bframes=5\", \n"
"                    \"VideoPreset\": \"veryslow\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 16.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ], \n"
"            \"Folder\": true, \n"
"            \"PresetName\": \"General\", \n"
"            \"Type\": 0\n"
"        }, \n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 128, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"Encode up to 3 minutes of video in large size for Discord Nitro Classic (50 MB or less) or up to 6 minutes of video in large size for Discord Nitro (100 MB or less). H.264 video (up to 1080p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Discord Nitro Large 3-6 Minutes 1080p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 2100, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"vbv-bufsize=21000:vbv-maxrate=2100:aq-mode=2:aq-strength=0.8:direct=auto:analyse=all:deblock=3,2\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 96, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"Encode up to 5 minutes of video in medium size for Discord Nitro Classic (50 MB or less) or up to 10 minutes of video in medium size for Discord Nitro (100 MB or less). H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Discord Nitro Medium 5-10 Minutes 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1250, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"vbv-bufsize=12500:vbv-maxrate=1250:aq-mode=2:aq-strength=0.8:direct=auto:analyse=all:deblock=3,2\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 96, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"Encode up to 10 minutes of video in small size for Discord Nitro Classic (50 MB or less) or up to 20 minutes of video in small size for Discord Nitro (100 MB or less). H.264 video (up to 480p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Discord Nitro Small 10-20 Minutes 480p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 550, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"vbv-bufsize=5500:vbv-maxrate=550:aq-mode=2:aq-strength=0.8:direct=auto:analyse=all:deblock=3,2\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 64, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"mono\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 360, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 480, \n"
"                    \"PresetDescription\": \"Encode up to 2 minutes of video in small size for Discord (8 MB or less). H.264 video (up to 360p30) and AAC mono audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Discord Small 2 Minutes 360p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 450, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"vbv-bufsize=4500:vbv-maxrate=450:aq-mode=2:aq-strength=0.8:direct=auto:analyse=all:deblock=3,2\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 32, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"mono\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 240, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 320, \n"
"                    \"PresetDescription\": \"Encode up to 5 minutes of video in tiny size for Discord (8 MB or less). H.264 video (up to 240p30) and AAC mono audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Discord Tiny 5 Minutes 240p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 165, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"vbv-bufsize=1650:vbv-maxrate=165:aq-mode=2:aq-strength=0.8:direct=auto:analyse=all:deblock=5,4:psy-rd=2.0:0.3:ref=5:bframes=5\", \n"
"                    \"VideoPreset\": \"veryslow\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 25.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 96, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"Encode up to 3 minutes of video in large size for Gmail (25 MB or less). H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Gmail Large 3 Minutes 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 900, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"vbv-bufsize=9000:vbv-maxrate=900:aq-mode=2:aq-strength=0.8:direct=auto:analyse=all:deblock=3,2\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 23.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 96, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"Encode up to 5 minutes of video in medium size for Gmail (25 MB or less). H.264 video (up to 480p30) and AAC stereo audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Gmail Medium 5 Minutes 480p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"vbv-bufsize=5000:vbv-maxrate=500:aq-mode=2:aq-strength=0.8:direct=auto:analyse=all:deblock=3,2\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 23.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 48, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"mono\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 288, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 384, \n"
"                    \"PresetDescription\": \"Encode up to 10 minutes of video in small size for Gmail (25 MB or less). H.264 video (up to 288p30) and AAC mono audio, in an MP4 container.\", \n"
"                    \"PresetName\": \"Gmail Small 10 Minutes 288p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 250, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"vbv-bufsize=2500:vbv-maxrate=250:aq-mode=2:aq-strength=0.8:direct=auto:analyse=all:deblock=3,2\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 23.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"48\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 2160, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 3840, \n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 2160p60) and high bit rate AAC stereo audio in an MP4 container. Compatible with video hosting services supporting 4K video, such as Vimeo and YouTube.\", \n"
"                    \"PresetName\": \"Vimeo YouTube HQ 2160p60 4K\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 18000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"5.2\", \n"
"                    \"VideoOptionExtra\": \"keyint=30:min-keyint=15:ref=1:bframes=2:b-pyramid=none\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"48\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1440, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 2560, \n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 1440p60) and high bit rate AAC stereo audio in an MP4 container. Compatible with video hosting services supporting 2.5K video, such as Vimeo and YouTube.\", \n"
"                    \"PresetName\": \"Vimeo YouTube HQ 1440p60 2.5K\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 12000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"5.2\", \n"
"                    \"VideoOptionExtra\": \"keyint=30:min-keyint=15:ref=1:bframes=2:b-pyramid=none\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"48\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 1080p60) and high bit rate AAC stereo audio in an MP4 container. Compatible with video hosting services supporting 1080p60, such as Vimeo and YouTube.\", \n"
"                    \"PresetName\": \"Vimeo YouTube HQ 1080p60\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 9000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.2\", \n"
"                    \"VideoOptionExtra\": \"keyint=30:min-keyint=15:ref=1:bframes=2:b-pyramid=none\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"48\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 720p60) and high bit rate AAC stereo audio in an MP4 container. Compatible with video hosting services supporting 720p60, such as Vimeo and YouTube.\", \n"
"                    \"PresetName\": \"Vimeo YouTube HQ 720p60\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 4500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.2\", \n"
"                    \"VideoOptionExtra\": \"keyint=30:min-keyint=15:ref=1:bframes=2:b-pyramid=none\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 19.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"48\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and high bit rate AAC stereo audio in an MP4 container. Compatible with most video hosting services, such as Vimeo and YouTube.\", \n"
"                    \"PresetName\": \"Vimeo YouTube 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 2500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.2\", \n"
"                    \"VideoOptionExtra\": \"keyint=30:min-keyint=15:ref=1:bframes=2:b-pyramid=none\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ], \n"
"            \"Folder\": true, \n"
"            \"PresetName\": \"Web\", \n"
"            \"Type\": 0\n"
"        }, \n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30) and AAC stereo audio, in an MP4 container. Compatible with Android devices.\", \n"
"                    \"PresetName\": \"Android 1080p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container. Compatible with Android devices.\", \n"
"                    \"PresetName\": \"Android 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 3000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 128, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 576, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"H.264 video (up to 576p25) and AAC stereo audio, in an MP4 container. Compatible with Android devices.\", \n"
"                    \"PresetName\": \"Android 576p25\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1800, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"25\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 128, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"H.264 video (up to 480p30) and AAC stereo audio, in an MP4 container. Compatible with Android devices.\", \n"
"                    \"PresetName\": \"Android 480p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 2160, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 3840, \n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Apple iPhone 7, 7 Plus, 8, 8 Plus, X, XR, XS, XS Max; Apple TV 4K.\", \n"
"                    \"PresetName\": \"Apple 2160p60 4K HEVC Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 8000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x265\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 24.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Apple iPad 5th and 6th Generation; iPad mini 2, 3, and 4; iPad Air 1st Generation and Air 2; iPad Pro 1st, 2nd, and 3rd Generation; Apple TV 4th Generation and later.\", \n"
"                    \"PresetName\": \"Apple 1080p60 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.2\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Apple iPhone 5, 5s, SE, 6, 6 Plus, 6s, 6s Plus, and later; iPod touch 6th Generation; iPad 3rd, 4th Generation and later; iPad mini 1st Generation and later; Apple TV 3rd, 4th Generation and later.\", \n"
"                    \"PresetName\": \"Apple 1080p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Apple iPhone 4, 4S, and later; iPod touch 4th, 5th Generation and later; iPad 1st Generation, iPad 2, and later; Apple TV 2nd Generation and later.\", \n"
"                    \"PresetName\": \"Apple 720p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 3000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 540, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 960, \n"
"                    \"PresetDescription\": \"H.264 video (up to 540p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Apple iPhone 1st Generation, 3G, 3GS, and later; iPod touch 1st, 2nd, 3rd Generation and later; iPod Classic; Apple TV 1st Generation and later.\", \n"
"                    \"PresetName\": \"Apple 540p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 2500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"qpmin=4:cabac=0:ref=2:b-pyramid=none:weightb=0:weightp=0:vbv-maxrate=9500:vbv-bufsize=9500\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": true, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 240, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 16, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 320, \n"
"                    \"PresetDescription\": \"H.264 video (up to 240p30) and AAC stereo audio, in an MP4 container. Compatible with Apple iPod 5th Generation and later.\", \n"
"                    \"PresetName\": \"Apple 240p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 350, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"1.3\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"baseline\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": true, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 2160, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 3840, \n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Google Chromecast Ultra.\", \n"
"                    \"PresetName\": \"Chromecast 2160p60 4K HEVC Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 8000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x265\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 24.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Google Chromecast 3rd Generation.\", \n"
"                    \"PresetName\": \"Chromecast 1080p60 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.2\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": true, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Google Chromecast 1st, 2nd Generation and later.\", \n"
"                    \"PresetName\": \"Chromecast 1080p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 2160, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 3840, \n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) audio, in an MP4 container. Compatible with Amazon Fire TV 2nd Generation and later; Fire TV Cube, Fire TV Stick 4K.\", \n"
"                    \"PresetName\": \"Amazon Fire 2160p60 4K HEVC Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 8000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x265\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 24.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) audio, in an MP4 container. Compatible with Amazon Fire TV 1st Generation and later; Fire TV Stick 1st Generation and later; Fire HD 10 7th Generation (2017); Fire HDX 4th Generation (2014).\", \n"
"                    \"PresetName\": \"Amazon Fire 1080p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container. Compatible with Amazon Fire HD 4th Generation (2014) and later; Kindle Fire HDX 3rd Generation (2013); Kindle Fire HD 2nd Generation (2012) and later.\", \n"
"                    \"PresetName\": \"Amazon Fire 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 3000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 2160, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 3840, \n"
"                    \"PresetDescription\": \"H.264 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Playstation 4 Pro.\", \n"
"                    \"PresetName\": \"Playstation 2160p60 4K Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 12000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"5.2\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 24.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Playstation 3 and 4.\", \n"
"                    \"PresetName\": \"Playstation 1080p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container. Compatible with Playstation Vita TV.\", \n"
"                    \"PresetName\": \"Playstation 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 3000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 540, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 960, \n"
"                    \"PresetDescription\": \"H.264 video (up to 540p30) and AAC stereo audio, in an MP4 container. Compatible with Playstation Vita.\", \n"
"                    \"PresetName\": \"Playstation 540p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 2250, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\", \n"
"                        \"copy:dts\", \n"
"                        \"copy:eac3\", \n"
"                        \"copy:mp3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 2160, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 3840, \n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60), AAC stereo audio, and surround audio, in an MKV container. Compatible with Roku 4, Streaming Stick+, Premiere+, and Ultra.\", \n"
"                    \"PresetName\": \"Roku 2160p60 4K HEVC Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 8000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x265\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 24.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Roku 1080p models.\", \n"
"                    \"PresetName\": \"Roku 1080p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Roku 720p models.\", \n"
"                    \"PresetName\": \"Roku 720p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 3000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 576, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"H.264 video (up to 576p25) and AAC stereo audio, in an MP4 container. Compatible with Roku standard definition models.\", \n"
"                    \"PresetName\": \"Roku 576p25\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1800, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"25\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": true, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"H.264 video (up to 480p30) and AAC stereo audio, in an MP4 container. Compatible with Roku standard definition models.\", \n"
"                    \"PresetName\": \"Roku 480p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Xbox One.\", \n"
"                    \"PresetName\": \"Xbox 1080p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }, \n"
"                        {\n"
"                            \"AudioBitrate\": 640, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"copy:ac3\", \n"
"                            \"AudioMixdown\": \"none\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": false, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 8, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Xbox 360.\", \n"
"                    \"PresetName\": \"Xbox Legacy 1080p30 Surround\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ], \n"
"            \"Folder\": true, \n"
"            \"PresetName\": \"Devices\", \n"
"            \"Type\": 0\n"
"        }, \n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 2160, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 3840, \n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60) and AAC stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"H.265 MKV 2160p60\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 8000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x265\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 24.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.265 video (up to 1080p30) and AAC stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"H.265 MKV 1080p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 2500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x265\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"H.265 video (up to 720p30) and AAC stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"H.265 MKV 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 2000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x265\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 576, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"H.265 video (up to 576p25) and AAC stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"H.265 MKV 576p25\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1200, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x265\", \n"
"                    \"VideoFramerate\": \"25\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"H.265 video (up to 480p30) and AAC stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"H.265 MKV 480p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x265\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1\", \n"
"                    \"VideoPreset\": \"slow\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 2160, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 3840, \n"
"                    \"PresetDescription\": \"H.264 video (up to 2160p60) and AAC stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"H.264 MKV 2160p60\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 12000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"5.2\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 24.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30) and AAC stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"H.264 MKV 1080p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"4.0\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and AAC stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"H.264 MKV 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 3000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 576, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"H.264 video (up to 576p25) and AAC stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"H.264 MKV 576p25\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1800, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"25\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"H.264 video (up to 480p30) and AAC stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"H.264 MKV 480p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"3.1\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"main\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [], \n"
"                    \"AudioEncoderFallback\": \"opus\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"opus\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 2160, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 3840, \n"
"                    \"PresetDescription\": \"VP9 video (up to 2160p60) and Opus stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"VP9 MKV 2160p60\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 8000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"vp9\", \n"
"                    \"VideoFramerate\": \"60\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"qmin=0:qmax=30\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"auto\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 24.0, \n"
"                    \"VideoQualityType\": 1, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [], \n"
"                    \"AudioEncoderFallback\": \"opus\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"opus\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"VP9 video (up to 1080p30) and Opus stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"VP9 MKV 1080p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 4500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"vp9\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"qmin=0:qmax=30\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"auto\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 1, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [], \n"
"                    \"AudioEncoderFallback\": \"opus\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"opus\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"VP9 video (up to 720p30) and Opus stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"VP9 MKV 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 2500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"vp9\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"qmin=0:qmax=30\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"auto\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 21.0, \n"
"                    \"VideoQualityType\": 1, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [], \n"
"                    \"AudioEncoderFallback\": \"opus\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"opus\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 576, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"VP9 video (up to 576p25) and Opus stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"VP9 MKV 576p25\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"vp9\", \n"
"                    \"VideoFramerate\": \"25\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"qmin=0:qmax=30\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"auto\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 1, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [], \n"
"                    \"AudioEncoderFallback\": \"opus\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"opus\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"VP9 video (up to 480p30) and Opus stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"VP9 MKV 480p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1250, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"vp9\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"qmin=0:qmax=30\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"auto\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 20.0, \n"
"                    \"VideoQualityType\": 1, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [], \n"
"                    \"AudioEncoderFallback\": \"vorbis\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"vorbis\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"VP8 video (up to 1080p30) and Vorbis stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"VP8 MKV 1080p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"vp8\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"qmin=0:qmax=24\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"auto\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 4.0, \n"
"                    \"VideoQualityType\": 1, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [], \n"
"                    \"AudioEncoderFallback\": \"vorbis\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"vorbis\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 720, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1280, \n"
"                    \"PresetDescription\": \"VP8 video (up to 720p30) and Vorbis stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"VP8 MKV 720p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 3000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"vp8\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"qmin=0:qmax=24\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"auto\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 4.0, \n"
"                    \"VideoQualityType\": 1, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [], \n"
"                    \"AudioEncoderFallback\": \"vorbis\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"vorbis\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 576, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"VP8 video (up to 576p25) and Vorbis stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"VP8 MKV 576p25\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1800, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"vp8\", \n"
"                    \"VideoFramerate\": \"25\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"qmin=0:qmax=24\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"auto\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 4.0, \n"
"                    \"VideoQualityType\": 1, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [], \n"
"                    \"AudioEncoderFallback\": \"vorbis\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"vorbis\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mkv\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"default\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"decomb\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 480, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"auto\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 720, \n"
"                    \"PresetDescription\": \"VP8 video (up to 480p30) and Vorbis stereo audio, in an MKV container.\", \n"
"                    \"PresetName\": \"VP8 MKV 480p30\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": true, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 1500, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"vp8\", \n"
"                    \"VideoFramerate\": \"30\", \n"
"                    \"VideoFramerateMode\": \"pfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"qmin=0:qmax=24\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"auto\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 4.0, \n"
"                    \"VideoQualityType\": 1, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ], \n"
"            \"Folder\": true, \n"
"            \"PresetName\": \"Matroska\", \n"
"            \"Type\": 0\n"
"        }, \n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"none\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": false, \n"
"                    \"AudioTrackSelectionBehavior\": \"all\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": false, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"off\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"off\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 0, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 0, \n"
"                    \"PresetDescription\": \"Maximum bit rate, constant frame rate H.264 video and high bit rate AAC stereo audio in an MP4 container. For professional use as an intermediate format for video editing. Creates gigantic files.\", \n"
"                    \"PresetName\": \"Production Max\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": false, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 120000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"auto\", \n"
"                    \"VideoFramerateMode\": \"cfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"keyint=12:min-keyint=1:ref=1:bframes=0:qcomp=0.8:aq-strength=0.5:dct-decimate=0:fast-pskip=0:deblock=-2,-2\", \n"
"                    \"VideoPreset\": \"fast\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 2.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"none\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": false, \n"
"                    \"AudioTrackSelectionBehavior\": \"all\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": false, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"off\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"off\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 0, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 0, \n"
"                    \"PresetDescription\": \"High bit rate, constant frame rate H.264 video and high bit rate AAC stereo audio in an MP4 container. For professional use as an intermediate format for video editing. Creates very large files.\", \n"
"                    \"PresetName\": \"Production Standard\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": false, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 60000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"auto\", \n"
"                    \"VideoFramerateMode\": \"cfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"keyint=12:min-keyint=1:ref=1:bframes=0:qcomp=0.8:aq-strength=0.5:dct-decimate=0:fast-pskip=0:deblock=-2,-2\", \n"
"                    \"VideoPreset\": \"fast\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 10.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"none\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": false, \n"
"                    \"AudioTrackSelectionBehavior\": \"all\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": false, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"off\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"off\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 1080, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 1920, \n"
"                    \"PresetDescription\": \"Intra-only, constant frame rate H.264 video (up to 1080p) and high bit rate AAC stereo audio in an MP4 container. For professional use as a low resolution proxy format for video editing.\", \n"
"                    \"PresetName\": \"Production Proxy 1080p\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": false, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 30000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"auto\", \n"
"                    \"VideoFramerateMode\": \"cfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"keyint=1:min-keyint=1:ref=1:bframes=0:qcomp=0.8:aq-strength=0.5\", \n"
"                    \"VideoPreset\": \"superfast\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 18.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }, \n"
"                {\n"
"                    \"AlignAVStart\": false, \n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"none\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"stereo\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": false, \n"
"                    \"AudioTrackSelectionBehavior\": \"all\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": false, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"InlineParameterSets\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": false, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"off\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"off\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"light\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 540, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"off\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 960, \n"
"                    \"PresetDescription\": \"Intra-only, constant frame rate H.264 video (up to 540p) and high bit rate AAC stereo audio in an MP4 container. For professional use as a low resolution proxy format for video editing.\", \n"
"                    \"PresetName\": \"Production Proxy 540p\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": false, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": true, \n"
"                    \"SubtitleBurnBehavior\": \"foreign\", \n"
"                    \"SubtitleBurnDVDSub\": true, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 1, \n"
"                    \"VideoAvgBitrate\": 15000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"auto\", \n"
"                    \"VideoFramerateMode\": \"cfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"keyint=1:min-keyint=1:ref=1:bframes=0:qcomp=0.8:aq-strength=0.5\", \n"
"                    \"VideoPreset\": \"superfast\", \n"
"                    \"VideoProfile\": \"high\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 16.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": true, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ], \n"
"            \"Folder\": true, \n"
"            \"PresetName\": \"Production\", \n"
"            \"Type\": 0\n"
"        }\n"
"    ], \n"
"    \"PresetCLIDefault\": [\n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\", \n"
"                        \"copy:ac3\", \n"
"                        \"copy:eac3\", \n"
"                        \"copy:dtshd\", \n"
"                        \"copy:dts\", \n"
"                        \"copy:mp3\", \n"
"                        \"copy:truehd\", \n"
"                        \"copy:flac\"\n"
"                    ], \n"
"                    \"AudioEncoderFallback\": \"aac\", \n"
"                    \"AudioLanguageList\": [], \n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 128, \n"
"                            \"AudioCompressionLevel\": -1.0, \n"
"                            \"AudioDitherMethod\": \"auto\", \n"
"                            \"AudioEncoder\": \"aac\", \n"
"                            \"AudioMixdown\": \"dpl2\", \n"
"                            \"AudioNormalizeMixLevel\": false, \n"
"                            \"AudioSamplerate\": \"auto\", \n"
"                            \"AudioTrackDRCSlider\": 0.0, \n"
"                            \"AudioTrackGainSlider\": 0.0, \n"
"                            \"AudioTrackQuality\": -1.0, \n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ], \n"
"                    \"AudioSecondaryEncoderMode\": true, \n"
"                    \"AudioTrackSelectionBehavior\": \"first\", \n"
"                    \"ChapterMarkers\": true, \n"
"                    \"ChildrenArray\": [], \n"
"                    \"Default\": true, \n"
"                    \"FileFormat\": \"mp4\", \n"
"                    \"Folder\": false, \n"
"                    \"FolderOpen\": false, \n"
"                    \"Mp4HttpOptimize\": false, \n"
"                    \"Mp4iPodCompatible\": false, \n"
"                    \"PictureAutoCrop\": true, \n"
"                    \"PictureBottomCrop\": 0, \n"
"                    \"PictureChromaSmoothCustom\": \"\", \n"
"                    \"PictureChromaSmoothPreset\": \"off\", \n"
"                    \"PictureChromaSmoothTune\": \"none\", \n"
"                    \"PictureCombDetectCustom\": \"\", \n"
"                    \"PictureCombDetectPreset\": \"off\", \n"
"                    \"PictureDARWidth\": 0, \n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"                    \"PictureDeblockPreset\": \"off\", \n"
"                    \"PictureDeblockTune\": \"medium\", \n"
"                    \"PictureDeinterlaceCustom\": \"\", \n"
"                    \"PictureDeinterlaceFilter\": \"off\", \n"
"                    \"PictureDeinterlacePreset\": \"default\", \n"
"                    \"PictureDenoiseCustom\": \"\", \n"
"                    \"PictureDenoiseFilter\": \"off\", \n"
"                    \"PictureDenoisePreset\": \"medium\", \n"
"                    \"PictureDenoiseTune\": \"none\", \n"
"                    \"PictureDetelecine\": \"off\", \n"
"                    \"PictureDetelecineCustom\": \"\", \n"
"                    \"PictureForceHeight\": 0, \n"
"                    \"PictureForceWidth\": 0, \n"
"                    \"PictureHeight\": 0, \n"
"                    \"PictureItuPAR\": false, \n"
"                    \"PictureKeepRatio\": true, \n"
"                    \"PictureLeftCrop\": 0, \n"
"                    \"PictureLooseCrop\": false, \n"
"                    \"PictureModulus\": 2, \n"
"                    \"PicturePAR\": \"loose\", \n"
"                    \"PicturePARHeight\": 720, \n"
"                    \"PicturePARWidth\": 853, \n"
"                    \"PictureRightCrop\": 0, \n"
"                    \"PictureRotate\": \"angle=0:hflip=0\", \n"
"                    \"PictureSharpenCustom\": \"\", \n"
"                    \"PictureSharpenFilter\": \"off\", \n"
"                    \"PictureSharpenPreset\": \"medium\", \n"
"                    \"PictureSharpenTune\": \"none\", \n"
"                    \"PictureTopCrop\": 0, \n"
"                    \"PictureWidth\": 0, \n"
"                    \"PresetDescription\": \"\", \n"
"                    \"PresetName\": \"CLI Default\", \n"
"                    \"SubtitleAddCC\": false, \n"
"                    \"SubtitleAddForeignAudioSearch\": false, \n"
"                    \"SubtitleAddForeignAudioSubtitle\": false, \n"
"                    \"SubtitleBurnBDSub\": false, \n"
"                    \"SubtitleBurnBehavior\": \"none\", \n"
"                    \"SubtitleBurnDVDSub\": false, \n"
"                    \"SubtitleLanguageList\": [], \n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"                    \"Type\": 0, \n"
"                    \"UsesPictureFilters\": true, \n"
"                    \"UsesPictureSettings\": 0, \n"
"                    \"VideoAvgBitrate\": 6000, \n"
"                    \"VideoColorMatrixCodeOverride\": 0, \n"
"                    \"VideoEncoder\": \"x264\", \n"
"                    \"VideoFramerate\": \"auto\", \n"
"                    \"VideoFramerateMode\": \"vfr\", \n"
"                    \"VideoGrayScale\": false, \n"
"                    \"VideoLevel\": \"auto\", \n"
"                    \"VideoOptionExtra\": \"\", \n"
"                    \"VideoPreset\": \"medium\", \n"
"                    \"VideoProfile\": \"auto\", \n"
"                    \"VideoQSVAsyncDepth\": 4, \n"
"                    \"VideoQSVDecode\": false, \n"
"                    \"VideoQualitySlider\": 22.0, \n"
"                    \"VideoQualityType\": 2, \n"
"                    \"VideoScaler\": \"swscale\", \n"
"                    \"VideoTune\": \"\", \n"
"                    \"VideoTurboTwoPass\": false, \n"
"                    \"VideoTwoPass\": false, \n"
"                    \"x264Option\": \"\", \n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ], \n"
"            \"Folder\": true, \n"
"            \"PresetName\": \"CLI Defaults\", \n"
"            \"Type\": 0\n"
"        }\n"
"    ], \n"
"    \"PresetTemplate\": {\n"
"        \"Preset\": {\n"
"            \"AlignAVStart\": false, \n"
"            \"AudioCopyMask\": [\n"
"                \"copy:aac\", \n"
"                \"copy:ac3\", \n"
"                \"copy:dts\", \n"
"                \"copy:dtshd\", \n"
"                \"copy:eac3\", \n"
"                \"copy:flac\", \n"
"                \"copy:mp3\", \n"
"                \"copy:truehd\"\n"
"            ], \n"
"            \"AudioEncoderFallback\": \"ac3\", \n"
"            \"AudioLanguageList\": [\n"
"                \"any\"\n"
"            ], \n"
"            \"AudioList\": [\n"
"                {\n"
"                    \"AudioBitrate\": 192, \n"
"                    \"AudioCompressionLevel\": -1.0, \n"
"                    \"AudioDitherMethod\": \"auto\", \n"
"                    \"AudioEncoder\": \"copy:ac3\", \n"
"                    \"AudioMixdown\": \"dpl2\", \n"
"                    \"AudioNormalizeMixLevel\": false, \n"
"                    \"AudioSamplerate\": \"auto\", \n"
"                    \"AudioTrackDRCSlider\": 0.0, \n"
"                    \"AudioTrackGainSlider\": 0.0, \n"
"                    \"AudioTrackQuality\": -1.0, \n"
"                    \"AudioTrackQualityEnable\": false\n"
"                }\n"
"            ], \n"
"            \"AudioSecondaryEncoderMode\": true, \n"
"            \"AudioTrackSelectionBehavior\": \"first\", \n"
"            \"ChapterMarkers\": true, \n"
"            \"ChildrenArray\": [], \n"
"            \"Default\": false, \n"
"            \"FileFormat\": \"mp4\", \n"
"            \"Folder\": false, \n"
"            \"FolderOpen\": false, \n"
"            \"InlineParameterSets\": false, \n"
"            \"Mp4HttpOptimize\": false, \n"
"            \"Mp4iPodCompatible\": false, \n"
"            \"PictureAutoCrop\": true, \n"
"            \"PictureBottomCrop\": 0, \n"
"            \"PictureChromaSmoothCustom\": \"\", \n"
"            \"PictureChromaSmoothPreset\": \"off\", \n"
"            \"PictureChromaSmoothTune\": \"none\", \n"
"            \"PictureCombDetectCustom\": \"\", \n"
"            \"PictureCombDetectPreset\": \"off\", \n"
"            \"PictureDARWidth\": 0, \n"
"            \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\", \n"
"            \"PictureDeblockPreset\": \"off\", \n"
"            \"PictureDeblockTune\": \"medium\", \n"
"            \"PictureDeinterlaceCustom\": \"\", \n"
"            \"PictureDeinterlaceFilter\": \"off\", \n"
"            \"PictureDeinterlacePreset\": \"default\", \n"
"            \"PictureDenoiseCustom\": \"\", \n"
"            \"PictureDenoiseFilter\": \"off\", \n"
"            \"PictureDenoisePreset\": \"medium\", \n"
"            \"PictureDenoiseTune\": \"none\", \n"
"            \"PictureDetelecine\": \"off\", \n"
"            \"PictureDetelecineCustom\": \"\", \n"
"            \"PictureForceHeight\": 0, \n"
"            \"PictureForceWidth\": 0, \n"
"            \"PictureHeight\": 0, \n"
"            \"PictureItuPAR\": false, \n"
"            \"PictureKeepRatio\": true, \n"
"            \"PictureLeftCrop\": 0, \n"
"            \"PictureLooseCrop\": false, \n"
"            \"PictureModulus\": 2, \n"
"            \"PicturePAR\": \"loose\", \n"
"            \"PicturePARHeight\": 720, \n"
"            \"PicturePARWidth\": 853, \n"
"            \"PictureRightCrop\": 0, \n"
"            \"PictureRotate\": \"angle=0:hflip=0\", \n"
"            \"PictureSharpenCustom\": \"\", \n"
"            \"PictureSharpenFilter\": \"off\", \n"
"            \"PictureSharpenPreset\": \"medium\", \n"
"            \"PictureSharpenTune\": \"none\", \n"
"            \"PictureTopCrop\": 0, \n"
"            \"PictureWidth\": 0, \n"
"            \"PresetDescription\": \"\", \n"
"            \"PresetName\": \"Name Missing\", \n"
"            \"SubtitleAddCC\": false, \n"
"            \"SubtitleAddForeignAudioSearch\": false, \n"
"            \"SubtitleAddForeignAudioSubtitle\": false, \n"
"            \"SubtitleBurnBDSub\": false, \n"
"            \"SubtitleBurnBehavior\": \"none\", \n"
"            \"SubtitleBurnDVDSub\": false, \n"
"            \"SubtitleLanguageList\": [], \n"
"            \"SubtitleTrackSelectionBehavior\": \"none\", \n"
"            \"Type\": 1, \n"
"            \"UsesPictureFilters\": true, \n"
"            \"UsesPictureSettings\": 2, \n"
"            \"VideoAvgBitrate\": 1800, \n"
"            \"VideoColorMatrixCodeOverride\": 0, \n"
"            \"VideoEncoder\": \"x264\", \n"
"            \"VideoFramerate\": \"auto\", \n"
"            \"VideoFramerateMode\": \"vfr\", \n"
"            \"VideoGrayScale\": false, \n"
"            \"VideoLevel\": \"auto\", \n"
"            \"VideoOptionExtra\": \"\", \n"
"            \"VideoPreset\": \"medium\", \n"
"            \"VideoProfile\": \"auto\", \n"
"            \"VideoQSVAsyncDepth\": 4, \n"
"            \"VideoQSVDecode\": false, \n"
"            \"VideoQualitySlider\": 20.0, \n"
"            \"VideoQualityType\": 2, \n"
"            \"VideoScaler\": \"swscale\", \n"
"            \"VideoTune\": \"\", \n"
"            \"VideoTurboTwoPass\": false, \n"
"            \"VideoTwoPass\": false, \n"
"            \"x264Option\": \"\", \n"
"            \"x264UseAdvancedOptions\": false\n"
"        }, \n"
"        \"VersionMajor\": 42, \n"
"        \"VersionMicro\": 0, \n"
"        \"VersionMinor\": 0\n"
"    }\n"
"}\n";
